from typing import List

from .. import document
from ..transport import GeminiUrl

DEMO_TEXT = """\
# Welcome to McRoss Browser

## List

* おぼえていますか　目と目が会った時を
* Do you remember? The time when our eyes first met?
* おぼえていますか　手と手触れ会った時
* Do you remember? The time when our hands first touched?

## Links

=> gemini://gemini.circumlunar.space/	Gemini homepage
=> gemini://gus.guru/     Gemini Universal Search engine
=> gemini://gemini.conman.org/test/torture/ 	Gemini client torture test

=> relative/ Relative link
=> /relative/ Relative link starting with "/"
=> https://lists.orbitalfox.eu/listinfo/gemini?foo=bar HTTP link

## Codes

```
[tool.poetry]
name = "mcross"
version = "0.1.0"
description = "Do you remember www?"
authors = ["nhanb <hi@imnhan.com>"]
license = "MIT"
```
"""


class History:
    h: List[GeminiUrl]
    current_index: int

    def __init__(self):
        self.h = []
        self.current_index = None

    def visit(self, url: GeminiUrl):
        # remove forward history first:
        if self.current_index is not None:
            self.h = self.h[: self.current_index + 1]
        self.h.append(url)
        self.current_index = len(self.h) - 1

    def go_back(self):
        if self.can_go_back():
            self.current_index -= 1

    def go_forward(self):
        if self.can_go_forward():
            self.current_index += 1

    def can_go_back(self):
        return self.current_index not in [None, 0]

    def can_go_forward(self):
        return self.current_index is not None and self.current_index < len(self.h) - 1

    def get_current_url(self):
        try:
            return self.h[self.current_index]
        except (IndexError, TypeError):
            return None


class Model:
    plaintext = ""
    gemini_nodes = None
    history: History

    def __init__(self):
        self.history = History()
        self.update_content(DEMO_TEXT)

    def update_content(self, plaintext):
        self.plaintext = plaintext
        self.gemini_nodes = []
        try:
            self.gemini_nodes = document.parse(plaintext)
        except Exception:
            print("Invalid gemini document!")



import logging
import traceback
from ssl import SSLCertVerificationError
from tkinter import TclError, Tk, messagebox

import curio

from ..transport import (
    GeminiUrl,
    NonAbsoluteUrlWithoutContextError,
    UnsupportedProtocolError,
    get,
)
from .model import Model
from .view import WAITING_CURSOR, View

statusbar_logger = logging.getLogger("statusbar")


class Controller:
    def __init__(self):
        self.root = Tk()
        self.model = Model()
        self.view = View(self.root, self.model)
        self.root.title("McRoss Browser")
        self.root.geometry("800x600")

        # Coroutine magic follows:

        self.pending_coros = []

        def schedule_as_coro(func):
            def do_schedule(*args):
                task = curio.spawn(
                    self.show_waiting_cursor_during_task(func, *args), daemon=True
                )
                self.pending_coros.append(task)

            return do_schedule

        self.view.go_callback = schedule_as_coro(self.go_callback)
        self.view.link_click_callback = schedule_as_coro(self.link_click_callback)
        self.view.back_callback = schedule_as_coro(self.back_callback)
        self.view.forward_callback = schedule_as_coro(self.forward_callback)

    def run(self):
        # Instead of running tkinter's root.mainloop() directly,
        # we rely on curio's event loop instead.
        # The main() coroutine does these things in an infinite loop:
        #   - do tk's necessary GUI with root.update()
        #   - run pending coroutines if there's any. This is used to run callbacks
        #     triggered by the view.
        #   - sleep a little so we don't loop root.update() too quickly.
        async def main():
            try:
                while True:
                    self.root.update()
                    for coroutine in self.pending_coros:
                        await coroutine
                    self.pending_coros = []
                    await curio.sleep(0.016)
                    # 16ms = 1/60 - we're targeting around 60fps
                    # Yes it's wasteful to call root.update() that fast.
                    # In practice CPU usage idles around 4% on my i5 but hey it's not
                    # spinning up my laptop fans yet.
                    # Doesn't seem like there's a better way atm. The alternative
                    # described at [1] is multithreading which I'm not a fan of.
                    # [1] https://github.com/dabeaz/curio/issues/111
            except TclError as e:
                if "application has been destroyed" not in str(e):
                    raise

        curio.run(main)

    async def show_waiting_cursor_during_task(self, func, *args):
        self.view.text.config(cursor=WAITING_CURSOR)
        self.root.config(cursor=WAITING_CURSOR)
        self.view.allow_changing_cursor = False

        try:
            await func(*args)
        except Exception:
            # a catch-all here so that our show_waiting...() coroutine can be yielded
            traceback.print_exc()

        # reset cursor to default values
        self.view.text.config(cursor="xterm")
        self.root.config(cursor="arrow")
        self.view.allow_changing_cursor = True

    async def go_callback(self, url: str):
        url = GeminiUrl.parse_absolute_url(url)
        await self.visit_link(url)

    async def link_click_callback(self, raw_url):
        try:
            url = GeminiUrl.parse(raw_url, self.model.history.get_current_url())
            await self.visit_link(url)
        except NonAbsoluteUrlWithoutContextError:
            messagebox.showwarning(
                "Ambiguous link",
                "Cannot visit relative urls without a current_url context",
            )
        except UnsupportedProtocolError as e:
            messagebox.showinfo(
                "Unsupported protocol", f"{e} links are unsupported (yet?)"
            )
        except SSLCertVerificationError:
            messagebox.showerror(
                "Invalid server certificate",
                "Server is NOT using a valid CA-approved TLS certificate.",
            )

    async def visit_link(self, url: GeminiUrl):
        try:
            resp = await self.load_page(url)
            self.model.history.visit(resp.url)
            self.view.render_page()
        except ConnectionError as e:
            statusbar_logger.info(str(e))
            raise

    async def back_callback(self):
        self.model.history.go_back()
        await self.load_page(self.model.history.get_current_url())
        self.view.render_page()

    async def forward_callback(self):
        self.model.history.go_forward()
        await self.load_page(self.model.history.get_current_url())
        self.view.render_page()

    async def load_page(self, url: GeminiUrl):
        statusbar_logger.info(f"Requesting {url}...")
        resp = await get(url)
        statusbar_logger.info(f"{resp.status} {resp.meta}")

        async def clear_status_bar_later():
            await curio.sleep(2)
            statusbar_logger.info("")

        await curio.spawn(clear_status_bar_later(), daemon=True)

        if resp.status.startswith("2"):
            self.model.update_content(resp.body.decode())
        else:
            self.model.update_content(
                "\n".join(
                    [
                        "Error:",
                        f"{resp.status} {resp.meta}",
                        resp.body.decode() if resp.body else "",
                    ]
                )
            )
        return resp



import re
from urllib.parse import urlparse

import curio

MAX_RESP_HEADER_BYTES = 2 + 1 + 1024 + 2  # <STATUS><whitespace><META><CR><LF>
MAX_RESP_BODY_BYTES = 1024 * 1024 * 5
MAX_REDIRECTS = 3


# Wanted to use a dataclass here but ofc it doesn't allow a slotted class to
# have fields with default values:
# https://stackoverflow.com/questions/50180735/how-can-dataclasses-be-made-to-work-better-with-slots
# Maaaaybe I should just use attrs and call it a day.
class Response:
    __slots__ = ("status", "meta", "body", "url")

    def __init__(self, status: str, meta: str, url, body: bytes = None):
        self.status = status
        self.meta = meta
        self.body = body
        self.url = url

    def __repr__(self):
        return f"Response(status={repr(self.status)}, meta={repr(self.meta)})"


def _parse_resp_header(header, pattern=re.compile(r"^(\d\d)\s+(.{,1024})\r\n$")):
    match = pattern.match(header)
    assert match is not None, f"Malformed response header: {header}"
    status = match.group(1)
    meta = match.group(2)
    return status, meta


# TODO: GeminiUrl's context-aware parse() method probably doesn't belong
# in a "transport" module.


class UnsupportedProtocolError(Exception):
    pass


class NonAbsoluteUrlWithoutContextError(Exception):
    pass


class GeminiUrl:
    PROTOCOL = "gemini"
    host: str
    port: int
    path: str

    def __init__(self, host, port, path):
        """
        You probably don't want to use this constructor directly.
        Use one of the parse methods instead.
        """
        self.host = host
        self.port = port
        self.path = path

    def __repr__(self):
        return f"{self.PROTOCOL}://{self.host}:{self.port}{self.path}"

    def without_protocol(self):
        if self.port == 1965:
            return f"{self.host}{self.path}"
        else:
            return f"{self.host}:{self.port}{self.path}"

    @classmethod
    def parse(cls, text, current_url):
        assert not re.search(r"\s", text), "Url should not contain any whitespace!"

        protocol = urlparse(text).scheme
        if protocol == cls.PROTOCOL:
            return cls.parse_absolute_url(text)

        if protocol:
            raise UnsupportedProtocolError(protocol)

        if current_url is None:
            raise NonAbsoluteUrlWithoutContextError(text)

        # relative url starting from top level
        if text.startswith("/"):
            return GeminiUrl(current_url.host, current_url.port, text)

        # just relative url:
        # trim stuff after the last `/` - for example:
        #   current url: gemini://example.com/foo/bar
        #   raw url text: yikes
        #   => parsed url: gemini://example.com/foo/yikes
        current_path = current_url.path[: current_url.path.rfind("/") + 1]
        return GeminiUrl(current_url.host, current_url.port, current_path + text)

    @staticmethod
    def parse_absolute_url(text):
        # TODO: urlparse doesn't seem that foolproof. Revisit later.
        parsed = urlparse(text)
        return GeminiUrl(parsed.hostname, parsed.port or 1965, parsed.path)


async def raw_get(url: GeminiUrl):
    sock = await curio.open_connection(
        url.host, url.port, ssl=True, server_hostname=url.host
    )
    async with sock:
        await sock.sendall(f"gemini://{url.host}{url.path}\r\n".encode())
        header = (await sock.recv(MAX_RESP_HEADER_BYTES)).decode()
        status, meta = _parse_resp_header(header)
        resp = Response(status=status, meta=meta, url=url)

        if status.startswith("2"):
            body = b""
            msg = await sock.recv(4096)
            body += msg
            while msg and len(body) <= MAX_RESP_BODY_BYTES:
                msg = await sock.recv(4096)
                body += msg
            resp.body = body

        return resp


async def get(url: GeminiUrl, redirect_count=0):
    resp = await raw_get(url)
    if resp.status.startswith("3") and redirect_count < MAX_REDIRECTS:
        redirect_count += 1
        new_url = GeminiUrl.parse_absolute_url(resp.meta)
        print(f"Redirecting to {new_url}, count={redirect_count}")
        return await get(new_url, redirect_count)
    return resp



import logging
import sys
from tkinter import Text, Tk, font, ttk

from ..document import (
    GeminiNode,
    H1Node,
    H2Node,
    H3Node,
    LinkNode,
    ListItemNode,
    PreformattedNode,
    TextNode,
)
from .model import Model
from .widgets import McEntry, ReadOnlyText

# OS-specific values
if sys.platform == "win32":
    TTK_THEME = "vista"
    POINTER_CURSOR = "center_ptr"
    WAITING_CURSOR = "wait"
elif sys.platform == "darwin":
    TTK_THEME = "aqua"
    POINTER_CURSOR = "pointinghand"
    WAITING_CURSOR = "wait"
else:
    TTK_THEME = "clam"
    POINTER_CURSOR = "hand1"
    WAITING_CURSOR = "watch"


def pick_font(names):
    available = font.families()
    picked = None
    for name in names:
        if name in available:
            picked = name
            break

    if not picked:
        picked = names[-1]

    print("Picked font:", picked)
    return picked


def register_status_bar_log_handler(status_bar: ttk.Label):
    class StatusBarHandler(logging.Handler):
        def emit(self, record):
            status_bar.config(text=self.format(record))

    logger = logging.getLogger("statusbar")
    logger.setLevel(logging.INFO)
    logger.addHandler(StatusBarHandler())


class View:
    model: Model
    address_bar: ttk.Entry
    go_btn: ttk.Button
    back_btn: ttk.Button
    forward_btn: ttk.Button
    text: Text
    status_bar: ttk.Label

    allow_changing_cursor = True

    go_callback = None
    link_click_callback = None
    back_callback = None
    forward_callback = None

    def __init__(self, root: Tk, model: Model):
        self.model = model

        # first row - address bar + buttons
        row1 = ttk.Frame(root)
        row1.pack(fill="x")

        # second row - web viewport
        row2 = ttk.Frame(root)
        row2.pack(fill="both", expand=True)

        # third row - status bar
        status_bar = ttk.Label(root)
        self.status_bar = status_bar
        status_bar.config(justify="right")
        status_bar.pack(fill="x")
        register_status_bar_log_handler(status_bar)

        # Back/Forward buttons
        back_btn = ttk.Button(
            row1, text="◀", width=3, command=lambda: self.back_callback()
        )
        forward_btn = ttk.Button(
            row1, text="▶", width=3, command=lambda: self.forward_callback()
        )
        back_btn.pack(side="left", padx=2)
        forward_btn.pack(side="left", padx=2)
        self.back_btn = back_btn
        self.forward_btn = forward_btn

        # Address bar prefix
        address_prefix = ttk.Label(row1, text="gemini://")
        address_prefix.pack(side="left")

        # Address bar
        address_bar = McEntry(row1)
        self.address_bar = address_bar
        address_bar.pack(side="left", fill="both", expand=True, padx=3, pady=3)
        address_bar.bind("<Return>", self._on_go)
        address_bar.bind("<KP_Enter>", self._on_go)
        address_bar.focus_set()

        # Go button
        go_btn = ttk.Button(row1, text="三三ᕕ( ᐛ )ᕗ", command=self._on_go, width=10)
        self.go_btn = go_btn
        go_btn.pack(side="left", pady=3)

        # Main viewport implemented as a Text widget.
        text = ReadOnlyText(row2)
        self.text = text
        self.render_page()
        text_font = pick_font(
            [
                "Charis SIL",
                "Source Serif Pro",
                "Cambria",
                "Georgia",
                "DejaVu Serif",
                "Times New Roman",
                "Times",
                "TkTextFont",
            ]
        )
        mono_font = pick_font(["Ubuntu Mono", "Consolas", "Courier", "TkFixedFont"])
        text.config(
            font=(text_font, 13),
            bg="#fff8dc",
            fg="black",
            padx=5,
            pady=5,
            # hide blinking insertion cursor:
            insertontime=0,
            # prevent text widget from pushing scrollbar/status bar out of the window:
            width=1,
            height=1,
        )
        text.pack(side="left", fill="both", expand=True)
        text.tag_config("link", foreground="brown")
        text.tag_bind("link", "<Enter>", self._on_link_enter)
        text.tag_bind("link", "<Leave>", self._on_link_leave)
        text.tag_bind("link", "<Button-1>", self._on_link_click)
        text.tag_config("pre", font=(mono_font, 13))
        text.tag_config("listitem", foreground="#044604")

        base_heading_font = font.Font(font=text["font"])
        base_heading_font.config(weight="bold")
        h1_font = font.Font(font=base_heading_font)
        h1_font.config(size=h1_font["size"] + 8)
        text.tag_config("h1", font=h1_font)
        h2_font = font.Font(font=base_heading_font)
        h2_font.config(size=h2_font["size"] + 4)
        text.tag_config("h2", font=h2_font)
        h3_font = font.Font(font=base_heading_font)
        text.tag_config("h3", font=h3_font)

        text_scrollbar = ttk.Scrollbar(row2, command=text.yview)
        text["yscrollcommand"] = text_scrollbar.set
        text_scrollbar.pack(side="left", fill="y")

        style = ttk.Style()
        style.theme_use(TTK_THEME)

    def _on_go(self, ev=None):
        if self.go_callback is not None:
            self.go_callback("gemini://" + self.address_bar.get())

    def _on_link_enter(self, ev):
        if self.allow_changing_cursor:
            self.text.config(cursor=POINTER_CURSOR)

    def _on_link_leave(self, ev):
        if self.allow_changing_cursor:
            self.text.config(cursor="xterm")

    def _on_link_click(self, ev):
        raw_url = get_content_from_tag_click_event(ev)
        self.link_click_callback(raw_url)

    def render_page(self):
        # Enable/Disable forward/back buttons according to history
        self.back_btn.config(
            state="normal" if self.model.history.can_go_back() else "disabled"
        )
        self.forward_btn.config(
            state="normal" if self.model.history.can_go_forward() else "disabled"
        )

        # Update url in address bar
        current_url = self.model.history.get_current_url()
        if current_url is not None:
            self.address_bar.delete(0, "end")
            self.address_bar.insert(0, current_url.without_protocol())

        # Update viewport
        self.text.delete("1.0", "end")
        if not self.model.gemini_nodes:
            self.text.insert("end", self.model.plaintext)
        else:
            for node in self.model.gemini_nodes:
                render_node(node, self.text)
            # delete final trailing newline:
            self.text.delete("insert-1c", "insert")


def render_node(node: GeminiNode, widget: Text):
    nodetype = type(node)
    if nodetype is TextNode:
        widget.insert("end", node.text)
    elif nodetype is LinkNode:
        widget.insert("end", "=> ")
        widget.insert("end", node.url, ("link",))
        if node.name:
            widget.insert("end", f" {node.name}")
    elif nodetype is PreformattedNode:
        widget.insert("end", f"```\n{node.text}\n```", ("pre",))
    elif nodetype is ListItemNode:
        widget.insert("end", node.text, ("listitem",))
    elif nodetype is H1Node:
        widget.insert("end", node.text, ("h1",))
    elif nodetype is H2Node:
        widget.insert("end", node.text, ("h2",))
    elif nodetype is H3Node:
        widget.insert("end", node.text, ("h3",))
    else:
        widget.insert("end", node.text)

    widget.insert("end", "\n")


def get_content_from_tag_click_event(event):
    # get the index of the mouse click
    index = event.widget.index("@%s,%s" % (event.x, event.y))

    # get the indices of all "link" tags
    tag_indices = list(event.widget.tag_ranges("link"))

    # iterate them pairwise (start and end index)
    for start, end in zip(tag_indices[0::2], tag_indices[1::2]):
        # check if the tag matches the mouse click index
        if event.widget.compare(start, "<=", index) and event.widget.compare(
            index, "<", end
        ):
            # return string between tag start and end
            return event.widget.get(start, end)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['mcross', 'mcross.gui']

package_data = \
{'': ['*']}

install_requires = \
['curio>=1.2,<2.0']

entry_points = \
{'console_scripts': ['mcross = mcross:run']}

setup_kwargs = {
    'name': 'mcross',
    'version': '0.5.4',
    'description': 'Do you remember www?',
    'long_description': None,
    'author': 'nhanb',
    'author_email': 'hi@imnhan.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import re

NEWLINE = "\n"
LINK_LINE_PATTERN = re.compile(r"^=>[ \t]+(\S+)([ \t]+(.+))?$")
HEADING_LINE_PATTERN = re.compile(r"^(#{1,3})\s+(.+)$")


class GeminiNode:
    text: str

    def __init__(self, text):
        self.text = text

    def __repr__(self):
        return f"{self.__class__.__name__}: {self.text.__repr__()}"


class TextNode(GeminiNode):
    pass


class ListItemNode(GeminiNode):
    pass


class H1Node(GeminiNode):
    pass


class H2Node(GeminiNode):
    pass


class H3Node(GeminiNode):
    pass


class LinkNode(GeminiNode):
    url: str
    name: str

    def __init__(self, text, url, name):
        self.text = text
        self.url = url
        self.name = name

    def __repr__(self):
        result = f"{self.__class__.__name__}: {self.url.__repr__()}"
        if self.name:
            result += f" {self.name.__repr__()}"
        return result


class PreformattedNode(GeminiNode):
    pass


def parse(text):
    """
    Naive one-pass parser.
    """
    nodes = []
    preformatted = None

    for line in text.strip().split(NEWLINE):

        if line == "```":
            if preformatted is None:
                # start preformatted mode
                preformatted = ""
            else:
                nodes.append(PreformattedNode(preformatted))
                preformatted = None

        elif preformatted is not None:
            if len(preformatted) > 0:
                preformatted += "\n"
            preformatted += line

        elif line.startswith("=> "):
            match = LINK_LINE_PATTERN.match(line)
            if not match:
                nodes.append(TextNode(line))
                continue
            url = match.group(1)
            name = match.group(3)  # may be None
            nodes.append(LinkNode(text=line, url=url, name=name))

        elif line.startswith("*"):
            nodes.append(ListItemNode(line))

        elif line.startswith("#"):
            match = HEADING_LINE_PATTERN.match(line)
            if not match:
                nodes.append(TextNode(line))
                continue
            # heading_text = match.group(2)  # not used yet
            hashes = match.group(1)
            level = len(hashes)
            if level == 1:
                nodes.append(H1Node(line))
            elif level == 2:
                nodes.append(H2Node(line))
            elif level == 3:
                nodes.append(H3Node(line))

        else:
            nodes.append(TextNode(line))

    return nodes



from idlelib.redirector import WidgetRedirector
from tkinter import Text, ttk


# Can't just use a Text widget with state='disabled' because that would
# also disable Ctrl+C / Ctrl+V.
# Further low-level reading:
# https://wiki.tcl-lang.org/page/Read%2Donly+text+widget
class ReadOnlyText(Text):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.redirector = WidgetRedirector(self)
        self.insert = self.redirector.register("insert", lambda *args, **kw: "break")
        self.delete = self.redirector.register("delete", lambda *args, **kw: "break")

        self.bind("<Control-a>", self._on_ctrl_a)

    def _on_ctrl_a(self, ev):
        self.tag_add("sel", "1.0", "end")
        return "break"


class McEntry(ttk.Entry):
    """
    Entry widget with reasonable defaults
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.bind("<Control-a>", self._on_ctrl_a)

    def _on_ctrl_a(self, ev):
        self.select_range(0, "end")
        self.icursor("end")
        return "break"
